# Copyright (C) 2013 Intra2net AG
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <http://www.gnu.org/licenses/lgpl-3.0.html>


import os
from tempfile import TemporaryDirectory

import deltatar.crypto as crypto
from deltatar.tarfile import TarFile, PAX_FORMAT, GNU_FORMAT, BLOCKSIZE
from . import BaseTest, new_volume_handler, closing_new_volume_handler
from .create_pseudo_random_files import create_file as create_random_file
from .test_multivol_compression_sizes import test as multivol_compr_test_func

class MultivolGnuFormatTest(BaseTest):
    """
    Test multivolume support in tarfile. Tar Format is specified at class level.
    """

    # used as the --format argument to tar command on tar file creation
    tar_command_format = "gnu"

    # used as Tarfile.open format option argument for tar file creation
    tarfile_format = GNU_FORMAT

    # overhead size used to calculate the exact maximum size of a tar file with
    # no extra volume that stores only one file. In case of GNU format this is
    # the size of three blocks:
    # * 1 block used to store the header information of the stored file
    # * 2 blocks used to mark the end of the tar file
    tarfile_overhead = 3*BLOCKSIZE
    file_overhead = 1*BLOCKSIZE

    # overhead size used to calculate the exact maximum size of a tar volume,
    # corresponding with a multivolume tar file storing a single file. In the
    # case of GNU format this is the same as tarfile_overhead.
    tarvol_overhead = 3*BLOCKSIZE

    def test_no_volume(self):
        """
        Create a tar file with only one file inside and no extra volumes
        """

        # create the content of the file to compress and hash it
        hash = self.create_file("big", 50000)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar", mode="w", format=self.tarfile_format)
        tarobj.add("big")
        tarobj.close()

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert not os.path.exists("sample.tar.1")

        os.unlink("big")
        assert not os.path.exists("big")

        # extract and check
        os.system("tar xfM sample.tar")
        assert os.path.exists("big")
        assert hash == self.md5sum("big")

    def test_volume_creation1(self):
        """
        Create a tar file with two volumes, only one file inside
        """

        # create the content of the file to compress and hash it
        hash = self.create_file("big", 50000)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar",
                              mode="w",
                              format=self.tarfile_format,
                              max_volume_size=30000,
                              new_volume_handler=new_volume_handler)
        tarobj.add("big")
        tarobj.close()

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert os.path.exists("sample.tar.1")
        assert not os.path.exists("sample.tar.2")

        os.unlink("big")
        assert not os.path.exists("big")

        # extract with normal tar and check output
        os.system("tar xfM sample.tar --file=sample.tar.1")
        assert os.path.exists("big")
        assert hash == self.md5sum("big")

    def test_volume_creation2(self):
        """
        Create a tar file with 2 extra volumes, only one file inside
        """

        # create the content of the file to compress and hash it
        hash = self.create_file("big", 50000)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar",
                              mode="w",
                              format=self.tarfile_format,
                              max_volume_size=20000,
                              new_volume_handler=new_volume_handler)
        tarobj.add("big")
        tarobj.close()

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert os.path.exists("sample.tar.1")
        assert os.path.exists("sample.tar.2")
        assert not os.path.exists("sample.tar.3")

        os.unlink("big")
        assert not os.path.exists("big")

        # extract with normal tar and check output
        os.system("tar xfM sample.tar --file=sample.tar.1 --file=sample.tar.2")
        assert os.path.exists("big")
        assert hash == self.md5sum("big")

    def test_multivol_multifiles(self):
        '''
        Create a tar file with two volumes and three files inside
        '''

        # create sample data
        hash = dict()
        hash["big"]  = self.create_file("big", 50000)
        hash["small"] = self.create_file("small", 100)
        hash["small2"] = self.create_file("small2", 354)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar",
                              mode="w",
                              format=self.tarfile_format,
                              max_volume_size=20000,
                              new_volume_handler=new_volume_handler)
        tarobj.add("big")
        tarobj.add("small")
        tarobj.add("small2")
        tarobj.close()

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert os.path.exists("sample.tar.1")
        assert os.path.exists("sample.tar.2")
        assert not os.path.exists("sample.tar.3")

        os.unlink("big")
        os.unlink("small")
        os.unlink("small2")

        # extract with normal tar and check output
        os.system("tar xfM sample.tar --file=sample.tar.1 --file=sample.tar.2")
        for key, value in hash.items():
            assert os.path.exists(key)
            assert value == self.md5sum(key)

    def test_get_file_size(self):
        '''
        Test _Stream.get_file_size which is the basis for multivol with
        compression.
        '''

        # create test files of different sizes
        size_factor = 4
        n_sizes = 13    # 1,4,16,64,256,1KiB=4**5 ... 1MiB=4**10 ... 1GiB=4**15
        next_size = 1
        sizes = []
        file_names = []
        for exponents in range(n_sizes):
            sizes.append(next_size)
            new_name = 'size_test_{:08d}'.format(next_size)
            file_names.append(new_name)
            self.create_file(new_name, next_size)
            next_size *= size_factor

        max_err = 0
        max_err_post = 0

        for mode, password in [ ('w|gz', None)  , ('w|bz2', None)
                              , ('w|xz', None)  , ('w#gz' , None)
                              , ('w#gz', "test"), ('w#tar', "test")
                              ]:
            tar_file_name = "size_test.tar." + mode[2:]
            for size_number in range(4,n_sizes):
                for order in 1,-1:   # small files first or big files first
                    encryptor = None
                    if password is not None:
                        encryptor = crypto.Encrypt (password=password, version=1,
                                                    paramversion=1)
                    tarobj = TarFile.open(tar_file_name,
                                          mode=mode,
                                          format=self.tarfile_format,
                                          encryption=encryptor)
                    for file_name in file_names[:size_number][::order]:
                        tarobj.add(file_name)
                    estimate = tarobj.fileobj.estim_file_size()
                    tarobj.close()
                    estimate_post = tarobj.fileobj.estim_file_size()
                    actual_size = os.stat(tar_file_name).st_size
                    err = abs(actual_size - estimate)
                    #print('mode {:>11s}, {:2} files (up to size {:9}): '
                    #      'estim={:9}, true={:9}, post={:9}, err={:5}'
                    #      .format(mode, size_number, sizes[size_number],
                    #              estimate, actual_size, estimate_post, err))
                    os.unlink(tar_file_name)
                    if err > max_err:
                        max_err = err
                    err = abs(actual_size - estimate_post)
                    if err > max_err_post:
                        max_err_post = err

        #print('max err is {}, post={}'.format(max_err, max_err_post))
        assert max_err < 13*1024
        assert max_err_post == 0

        # clean up
        for file_name in file_names:
            os.unlink(file_name)

    def test_multivol_compress_warning(self):
        """ check warning being issued if compressing multivolume with w: """
        with self.assertWarns(UserWarning):
            tarobj = TarFile.open("sample.tar.gz",
                                  mode="w:gz",
                                  format=self.tarfile_format,
                                  max_volume_size=30000,
                                  new_volume_handler=new_volume_handler)
            tarobj.close()

    def test_compress_single(self):
        ''' check creation of single volume when compression is on '''

        # create the content of the file to compress and hash it
        hash = self.create_file("big", 50000)

        # create the tar file with volumes and compression
        tarobj = TarFile.open("sample.tar.gz",
                              mode="w#gz",
                              format=self.tarfile_format,
                              max_volume_size=30000,
                              new_volume_handler=new_volume_handler,
                              debug=3)
        tarobj.add("big")
        tarobj.list()
        tarobj.close()

        # data fits into a single volume -- check that no second is created
        assert os.path.exists("sample.tar.gz")
        assert not os.path.exists("sample.tar.gz.1")

        # check size of first volume
        size = os.stat("sample.tar.gz").st_size
        assert 330 < size < 410, 'size of sample.tar.gz is {}'.format(size)

        os.unlink("big")
        assert not os.path.exists("big")

        # extract with normal tar and check output
        #print('unpacking:')
        import subprocess
        output = subprocess.check_output(
            "tar xvf sample.tar.gz".split(),
            universal_newlines=True)
        #for line in output.splitlines():
        #    print(line.rstrip())
        assert os.path.exists("big")
        assert hash == self.md5sum("big")

        os.unlink('big')

    def test_multivol_compress(self):
        ''' check creation of multiple volumes when compression is on '''

        # create a random file that is not so easy to compress...
        filename = create_random_file('./', 100000)
        hash = self.md5sum(filename)

        # need own volume handler so files maintain their suffix for gunzip
        def my_volume_handler(tarobj, base_name, volume_number):
            if not base_name.endswith('.tar.gz'):
                raise ValueError('need .tar.gz file!')
            tarobj.fileobj.close()
            new_name = '{}.{}.tar.gz'.format(base_name[:-7], volume_number)
            tarobj.open_volume(new_name)

        # create the tar file with volumes and compression
        tarobj = TarFile.open("sample.tar.gz",
                              mode="w#gz",
                              format=self.tarfile_format,
                              max_volume_size=30000,
                              new_volume_handler=my_volume_handler,
                              debug=3)
        tarobj.add(filename)
        tarobj.list()
        tarobj.close()

        # data fits into 2 volumes -- check that no third is created
        assert os.path.exists("sample.tar.gz")
        assert os.path.exists("sample.1.tar.gz")
        assert not os.path.exists("sample.tar.gz.1")
        assert not os.path.exists("sample.tar.gz.2")
        assert not os.path.exists("sample.2.tar.gz.2")

        os.unlink(filename)
        assert not os.path.exists(filename)

        # extract with shell means; slightly complicated because the linux
        # tar/gunzip cannot do gzipped-multi-volume archives
        #print('unpacking:')
        import subprocess
        for cmd in 'gunzip -v sample.tar.gz', 'gunzip -v sample.1.tar.gz', \
                'tar xvfM sample.tar --file=sample.1.tar':
            #print(cmd)
            output = subprocess.check_output(cmd.split(),
                                             universal_newlines=True)
            #for line in output.splitlines():
            #    print(line.rstrip())
        assert os.path.exists(filename)
        assert hash == self.md5sum(filename)

        os.unlink('sample.tar')
        os.unlink('sample.1.tar')
        os.unlink(filename)

    def test_volume_extract1(self):
        '''
        Create a tar file with multiple volumes and one file and extract it
        '''
        # create the content of the file to compress and hash it
        hash = self.create_file("big", 5*1024*1024)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar",
                              mode="w",
                              format=self.tarfile_format,
                              max_volume_size=3*1024*1024,
                              new_volume_handler=new_volume_handler)
        tarobj.add("big")
        tarobj.close()

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert os.path.exists("sample.tar.1")
        assert not os.path.exists("sample.tar.2")

        os.unlink("big")
        assert not os.path.exists("big")

        # extract and check output
        tarobj = TarFile.open("sample.tar",
                              mode="r",
                              new_volume_handler=new_volume_handler)
        tarobj.extractall()
        tarobj.close()
        assert os.path.exists("big")
        assert hash == self.md5sum("big")

    def test_volume_extract2(self):
        '''
        Create a multivolume tar file with gnu tar command, extract it with
        tarfile library
        '''
        # create the content of the file to compress and hash it
        hash = self.create_file("big", 5*1024*1024)

        # create the tar file with volumes
        os.system("tar cM --format=%s -L 3000 big --file=sample.tar "\
            "--file=sample.tar.1" % self.tar_command_format)

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert os.path.exists("sample.tar.1")
        assert not os.path.exists("sample.tar.2")

        os.unlink("big")
        assert not os.path.exists("big")

        # extract and check output
        tarobj = TarFile.open("sample.tar",
                              mode="r",
                              new_volume_handler=new_volume_handler)
        tarobj.extractall()
        tarobj.close()
        assert os.path.exists("big")
        assert hash == self.md5sum("big")

    def test_volume_extract3(self):
        '''
        Create a multivolume tar file with gnu tar command with multiple
        files, extract it with tarfile library
        '''
        # create the content of the file to compress and hash it
        hash = dict()
        hash["big"] = self.create_file("big", 5*1024*1024)
        hash["small"] = self.create_file("small", 100)
        hash["small2"] = self.create_file("small2", 354)

        # create the tar file with volumes
        os.system("tar cM --format=%s -L 3000 big small small2 --file=sample.tar "\
            "--file=sample.tar.1" % self.tar_command_format)

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert os.path.exists("sample.tar.1")
        assert not os.path.exists("sample.tar.2")

        for key, value in hash.items():
            os.unlink(key)
            assert not os.path.exists(key)

        # extract and check output
        tarobj = TarFile.open("sample.tar",
                              mode="r",
                              new_volume_handler=new_volume_handler)
        tarobj.extractall()
        tarobj.close()

        for key, value in hash.items():
            assert os.path.exists(key)
            assert value == self.md5sum(key)

    def test_multivol_multifile_extract(self):
        '''
        create a multivolume tar file with multiple files and extracts it
        '''

        # create sample data
        hash = dict()
        hash["big"] = self.create_file("big", 50000)
        hash["small"] = self.create_file("small", 100)
        hash["small2"] = self.create_file("small2", 354)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar",
                              mode="w",
                              format=self.tarfile_format,
                              max_volume_size=20000,
                              new_volume_handler=new_volume_handler)
        tarobj.add("big")
        tarobj.add("small")
        tarobj.add("small2")
        tarobj.close()

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert os.path.exists("sample.tar.1")
        assert os.path.exists("sample.tar.2")
        assert not os.path.exists("sample.tar.3")

        os.unlink("big")
        os.unlink("small")
        os.unlink("small2")

        # extract and check output
        tarobj = TarFile.open("sample.tar",
                              mode="r",
                              new_volume_handler=new_volume_handler)
        tarobj.extractall()
        tarobj.close()

        for key, value in hash.items():
            assert os.path.exists(key)
            assert value == self.md5sum(key)

    def test_multiple_files_extract(self):
        '''
        creates a simple tar file with no volumes and with multiple files
        inside and extracts it
        '''

        # create sample data
        hash = dict()
        hash["big"] = self.create_file("big", 50000)
        hash["small"] = self.create_file("small", 100)
        hash["small2"] = self.create_file("small2", 354)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar",
                              format=self.tarfile_format,
                              mode="w")
        tarobj.add("big")
        tarobj.add("small")
        tarobj.add("small2")
        tarobj.close()

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert not os.path.exists("sample.tar.1")

        os.unlink("big")
        os.unlink("small")
        os.unlink("small2")

        # extract and check output
        tarobj = TarFile.open("sample.tar",
                              mode="r",
                              new_volume_handler=new_volume_handler)
        tarobj.extractall()
        tarobj.close()

        for key, value in hash.items():
            assert os.path.exists(key)
            assert value == self.md5sum(key)

    def test_corner_case_split_size1(self):
        '''
        Creates a tar file with a single file inside that contains the maximum
        size allowed in one volume.
        '''
        hash = self.create_file("big", 5*1024*1024)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar",
                              mode="w",
                              format=self.tarfile_format,
                              # see tarfile_overhead description for details
                              max_volume_size=5*1024*1024 + self.tarfile_overhead,
                              new_volume_handler=new_volume_handler)
        tarobj.add("big")
        tarobj.close()

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert not os.path.exists("sample.tar.1")

        os.unlink("big")
        assert not os.path.exists("big")

        # extract and check output
        tarobj = TarFile.open("sample.tar",
                              mode="r",
                              new_volume_handler=new_volume_handler)
        tarobj.extractall()
        tarobj.close()
        assert os.path.exists("big")
        assert hash == self.md5sum("big")

    def test_corner_case_split_size2(self):
        '''
        Creates a tar file with a single file inside that contains the maximum
        size allowed in one volume.
        '''
        hash = self.create_file("big", 4*1024*1024)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar",
                              mode="w",
                              format=self.tarfile_format,
                              # see tarvol_overhead description for details
                              max_volume_size=2*1024*1024 + self.tarvol_overhead,
                              new_volume_handler=new_volume_handler)
        tarobj.add("big")
        tarobj.close()

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert os.path.exists("sample.tar.1")
        assert not os.path.exists("sample.tar.2")

        os.unlink("big")
        assert not os.path.exists("big")

        # extract and check output
        tarobj = TarFile.open("sample.tar",
                              mode="r",
                              new_volume_handler=new_volume_handler)
        tarobj.extractall()
        tarobj.close()
        assert os.path.exists("big")
        assert hash == self.md5sum("big")

    def test_corner_case_split_size3(self):
        '''
        Creates a tar file with a single file inside that contains the maximum
        size allowed in one volume but without the overhead.
        '''
        hash = self.create_file("big", 4*1024*1024)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar",
                              mode="w",
                              format=self.tarfile_format,
                              max_volume_size=2*1024*1024,
                              new_volume_handler=new_volume_handler)
        tarobj.add("big")
        tarobj.close()

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert os.path.exists("sample.tar.1")
        assert os.path.exists("sample.tar.2")
        assert not os.path.exists("sample.tar.3")

        os.unlink("big")
        assert not os.path.exists("big")

        # extract and check output
        tarobj = TarFile.open("sample.tar",
                              mode="r",
                              new_volume_handler=new_volume_handler)
        tarobj.extractall()
        tarobj.close()
        assert os.path.exists("big")
        assert hash == self.md5sum("big")

    def test_corner_case_split_size4(self):
        '''
        Creates a tar file with multiple files inside that contains the maximum
        size allowed in one volume.
        '''
        hash = dict()
        hash['big'] = self.create_file("big", 3*1024*1024)
        hash['small'] = self.create_file("small", 1*1024*1024)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar",
                              mode="w",
                              format=self.tarfile_format,
                              max_volume_size=(4*1024*1024 +
                                               self.tarfile_overhead +
                                               self.file_overhead),
                              new_volume_handler=new_volume_handler)
        tarobj.add("big")
        tarobj.add("small")
        tarobj.close()

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert not os.path.exists("sample.tar.1")

        for key, value in hash.items():
            os.unlink(key)
            assert not os.path.exists(key)

        # extract and check output
        tarobj = TarFile.open("sample.tar",
                              mode="r",
                              new_volume_handler=new_volume_handler)
        tarobj.extractall()
        tarobj.close()

        for key, value in hash.items():
            assert os.path.exists(key)
            assert value == self.md5sum(key)

    def test_corner_case_split_size5(self):
        '''
        Creates a tar file with multiple files inside that contains the maximum
        size allowed in one volume.
        '''
        hash = dict()
        hash['big'] = self.create_file("big", 3*1024*1024)
        hash['small'] = self.create_file("small", 1*1024*1024)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar",
                              mode="w",
                              format=self.tarfile_format,
                              max_volume_size=(2*1024*1024 +
                                               self.tarfile_overhead +
                                               self.file_overhead),
                              new_volume_handler=new_volume_handler)
        tarobj.add("big")
        tarobj.add("small")
        tarobj.close()

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert os.path.exists("sample.tar.1")
        assert not os.path.exists("sample.tar.2")

        for key, value in hash.items():
            os.unlink(key)
            assert not os.path.exists(key)

        # extract and check output
        tarobj = TarFile.open("sample.tar",
                              mode="r",
                              new_volume_handler=new_volume_handler)
        tarobj.extractall()
        tarobj.close()

        for key, value in hash.items():
            assert os.path.exists(key)
            assert value == self.md5sum(key)

    def test_volume_not_found(self):
        '''
        Create a tar file with multiple volumes and one file and extract it, but
        one of the volumes is missing
        '''
        # create the content of the file to compress and hash it
        hash = self.create_file("big", 5*1024*1024)

        # create the tar file with volumes
        tarobj = TarFile.open("sample.tar",
                              mode="w",
                              format=self.tarfile_format,
                              max_volume_size=2*1024*1024,
                              new_volume_handler=closing_new_volume_handler)
        tarobj.add("big")
        tarobj.close()

        # check that the tar volumes were correctly created
        assert os.path.exists("sample.tar")
        assert os.path.exists("sample.tar.1")
        assert os.path.exists("sample.tar.2")
        assert not os.path.exists("sample.tar.3")

        os.unlink("big")
        os.unlink("sample.tar.2")

        class VolumeNotFound(Exception):
            pass

        def new_volume_handler2(tarobj, base_name, volume_number):
            '''
            Handles the new volumes
            '''
            volume_path = "%s.%d" % (base_name, volume_number)

            try:
                tarobj.fileobj.close()
                tarobj.open_volume(volume_path)
            except OSError as e:
                # only volume number 2 is missing
                assert volume_number == 2
                raise e

        # extract and check output
        tarobj = TarFile.open("sample.tar",
                              mode="r",
                              new_volume_handler=new_volume_handler2)
        try:
            tarobj.extractall()
        except OSError as e:
            pass
        tarobj.close()
        assert os.path.exists("big")
        assert hash != self.md5sum("big")

    def test_multivol_compress_vol_size(self):
        """ test size of compressed volumes using "external" test routine

        created an extensive test of this in extra file, run here just 2 short
        versions
        """
        # params
        vol_size = 3   # MiB
        input_size_factor = 3   # --> add 3*3 MiB of data
        modes = ('w#gz', None), ('w#gz', "test1234")
        debug_level = 0   # no debug output
        clean_up_if_error = True   # leave no files behind

        with TemporaryDirectory(prefix='deltatar_test_multivol_') \
                as temp_dir:     # is deleted automatically after test
            for mode, password in modes:
                multivol_compr_test_func(vol_size, input_size_factor, mode,
                                         password,
                                         temp_dir,
                                         debug_level=debug_level,
                                         clean_up_if_error=clean_up_if_error)


class MultivolPaxFormatTest(MultivolGnuFormatTest):
    """
    Test multivolume support in tarfile with PAX format
    """

    tar_command_format = "pax"

    tarfile_format = PAX_FORMAT

    # overhead size used to calculate the exact maximum size of a tar file with
    # no extra volume that stores only one file. In case of GNU format this is
    # the size of three blocks:
    # * 1 block used to store the header information of the stored file
    # * 1 block used to store the header information of the pax header
    # * 1 block used to store the pax header
    # * 2 blocks used to mark the end of the tar file
    tarfile_overhead = 5*BLOCKSIZE
    file_overhead = 3*BLOCKSIZE

    # overhead size used to calculate the exact maximum size of a tar volume,
    # corresponding with a multivolume tar file storing a single file. In the
    # case of Pax format, it's the same as tarfile_overhead plus a block for
    # the global header
    tarvol_overhead = 6*BLOCKSIZE
